
module Examples.Edit where

import Control.Monad ((>=>),msum)
import Data.Generics (Data,mkQ,mkT)
import Data.Generics.Zipper -- requires, from Hackage: syz
import Data.Maybe (fromJust,fromMaybe)
import qualified Data.Set as Set (member)

import CC.Syntax
import CC.Static
import CC.Zipper

import CC.Semantics
-- import Examples.List
-- import Examples.Names


----------------------------------
-- Variation Editing Operations --
----------------------------------

--
-- dimension & choice manipulations
--

-- given an n-ary choice, produce an (n+1)-ary choice where the ith alternative
-- has been copied and appended to the end. i is indexed from 1.
copyAlt :: Int -> V a -> V a
copyAlt i (Chc d as) = Chc d (as ++ [as !! (i-1)])

-- add a new alternative to the end of a choice
addAlt :: V a -> V a -> V a
addAlt a (Chc d as) = Chc d (as ++ [a])


--
-- query operations
--

-- get the tags associated with a dimension.
getTags :: Data a => Dim -> V a -> Maybe [Tag]
getTags d (Dim d' ts e) | d == d' = Just ts
getTags d e = msum (ccQ (getTags d) e)



--------------------------------
-- Generic Editing Operations --
--------------------------------

-- context type (is a synonym for the Zipper (V a), see Zipper.hs)
--
type C a = Z a

-- editing components
--
type Locator a  = V a -> Maybe (C a)
type Splitter a = V a -> Maybe (C a,V a)
type Transf a   = V a -> V a
type Edit a     = C a -> C a
type Pred a     = V a -> Bool
-- type GSplitter a b = V a -> Maybe (C a,b)
-- type Splitters a = V a -> Maybe (C a,[V a])


--
-- auxiliary functions
--

-- catching errors in transformations
withFallback :: a -> Maybe a -> a
withFallback = fromMaybe


--
-- functions for finding and extracting expressions
--

-- find an expression that satisfies a predicate
find ::  Data a => Pred a -> Locator a
find p = match p . toZipper

-- find and expose an expression that satisfies a predicate
extract :: Data a => Pred a -> Splitter a
extract p e = do
    c <- find p e
    h <- getHole c
    return (c,h)

-- predicates and function for finding dimensions and choices
dimDef :: Dim -> Pred a
dimDef d (Dim d' _ _) = d == d'
dimDef _ _            = False

chcFor :: Dim -> Pred a
chcFor d (Chc d' _) = d == d'
chcFor _ _          = False

extractDim :: Data a => Dim -> Splitter a
extractDim = extract . dimDef

extractChc :: Data a => Dim -> Splitter a
extractChc = extract . chcFor



--
-- basic context manipulations
--

-- filling the hole of a context with an expression
infixr <@
(<@) :: Data a => C a -> V a -> V a
c <@ e = fromZipper (setHole e c)

-- turn an edit into a transformation
runEdit :: Data a => Edit a -> Transf a
runEdit f = fromZipper . f . toZipper

-- navigate a transformation to nodes enclosed by two conditions
editBetween :: Data a => (Pred a) -> (Pred a) -> Transf a -> Edit a
editBetween b e t c | atX b c = continue (trans (mkT t) c)
                    | atX e c = leftT edit c
                    | otherwise = continue c
  where continue = leftT edit . downT edit
        edit = editBetween b e t



--
-- restructuring variational values
--

-- lift a dimension to the top level
hoist :: Data a => Dim -> V a -> V a
hoist d e = withFallback e $ do
    (c,Dim _ ts e') <- extractDim d e
    return (Dim d ts (c <@ e'))

-- lift dimension safely without choice capture
safeHoist :: Data a => Dim -> V a -> V a
safeHoist d e = withFallback e $ do
    (c,Dim _ ts e') <- extractDim d e
    if d `Set.member` freeDims e then
       Nothing
    else
       return (Dim d ts (c <@ e'))

-- choice prioritization, assumes nested choice sits in 2nd alternative
prioritize :: Data a => Dim -> Dim -> V a -> V a
prioritize b a e = withFallback e $ do
    (dA,ae)            <- extractDim a e
    (cA,Chc _ [a1,a2]) <- extractChc a ae
    (cB,Chc _ [b1,b2]) <- extractChc b a2
    return $ dA <@ (Chc b [cB <@ b1,cA <@ (Chc a [a1,cB <@ b2])])

-- swapping dependency and nesting for two dimensions
invert :: Data a => Dim -> Dim -> V a -> V a
invert b a = prioritize b a . hoist b

-- apply a transformation to nodes enclosed by two conditions
inRange :: Data a => (V a -> V a) -> (Pred a,Pred a) -> V a -> V a
inRange f (begin,end) e = runEdit (editBetween begin end f) e

-- extend a dimension with a new tag, transforming all bound choices
extend :: Data a => Dim -> Tag -> (V a -> V a) -> V a -> V a
extend d t f e = withFallback e $ do
    (c, Dim _ ts e) <- extractDim d e
    let e' = f `inRange` (chcFor d,dimDef d) $ e
    return (c <@ Dim d (ts++[t]) e')

